@@ -1,11 +1,22 @@
-from tests.response import Response
-from uvicorn.protocols.http.h11_impl import H11Protocol
-from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol
-from uvicorn.protocols.websockets.wsproto_impl import WSProtocol
 import asyncio
+import logging
+
 import h11
 import pytest
 
+from tests.response import Response
+from uvicorn.config import Config
+from uvicorn.main import ServerState
+from uvicorn.protocols.http.h11_impl import H11Protocol
+from uvicorn.protocols.websockets.wsproto_impl import WSProtocol
+
+try:
+    from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol
+except ImportError:  # pragma: nocover
+    HttpToolsProtocol = None
+
+
+HTTP_PROTOCOLS = [p for p in [H11Protocol, HttpToolsProtocol] if p is not None]
 
 SIMPLE_GET_REQUEST = b"\r\n".join([b"GET / HTTP/1.1", b"Host: example.org", b"", b""])
 
@@ -48,14 +59,20 @@ FINISH_POST_REQUEST = b'{"hello": "world"}'
 
 HTTP10_GET_REQUEST = b"\r\n".join([b"GET / HTTP/1.0", b"Host: example.org", b"", b""])
 
-UPGRADE_REQUEST = b"\r\n".join([
-    b"GET / HTTP/1.1",
-    b"Host: example.org",
-    b"Connection: upgrade",
-    b"Upgrade: websocket",
-    b"",
-    b""
-])
+GET_REQUEST_WITH_RAW_PATH = b"\r\n".join(
+    [b"GET /one%2Ftwo HTTP/1.1", b"Host: example.org", b"", b""]
+)
+
+UPGRADE_REQUEST = b"\r\n".join(
+    [
+        b"GET / HTTP/1.1",
+        b"Host: example.org",
+        b"Connection: upgrade",
+        b"Upgrade: websocket",
+        b"",
+        b"",
+    ]
+)
 
 
 class MockTransport:
@@ -72,7 +89,7 @@ class MockTransport:
             "sockname": self.sockname,
             "peername": self.peername,
             "sslcontext": self.sslcontext,
-        }[key]
+        }.get(key)
 
     def write(self, data):
         assert not self.closed
@@ -132,91 +149,107 @@ class MockTask:
 def get_connected_protocol(app, protocol_cls, **kwargs):
     loop = MockLoop()
     transport = MockTransport()
-    protocol = protocol_cls(app, loop, **kwargs)
+    config = Config(app=app, **kwargs)
+    server_state = ServerState()
+    protocol = protocol_cls(config=config, server_state=server_state, _loop=loop)
     protocol.connection_made(transport)
     return protocol
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_get_request(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert b"Hello, world" in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("path", ["/", "/?foo", "/?foo=bar", "/?foo=bar&baz=1"])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
+def test_request_logging(path, protocol_cls, caplog):
+    get_request_with_query_string = b"\r\n".join(
+        ["GET {} HTTP/1.1".format(path).encode("ascii"), b"Host: example.org", b"", b""]
+    )
+    caplog.set_level(logging.INFO, logger="uvicorn.access")
+    logging.getLogger("uvicorn.access").propagate = True
+
+    app = Response("Hello, world", media_type="text/plain")
+
+    protocol = get_connected_protocol(app, protocol_cls, log_config=None)
+    protocol.data_received(get_request_with_query_string)
+    protocol.loop.run_one()
+
+    assert '"GET {} HTTP/1.1" 200'.format(path) in caplog.records[0].message
+
+
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_head_request(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_HEAD_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert b"Hello, world" not in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_post_request(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            self.scope = scope
+    async def app(scope, receive, send):
+        body = b""
+        more_body = True
+        while more_body:
+            message = await receive()
+            body += message.get("body", b"")
+            more_body = message.get("more_body", False)
+        response = Response(b"Body: " + body, media_type="text/plain")
+        await response(scope, receive, send)
 
-        async def __call__(self, receive, send):
-            body = b""
-            more_body = True
-            while more_body:
-                message = await receive()
-                body += message.get("body", b"")
-                more_body = message.get("more_body", False)
-            response = Response(b"Body: " + body, media_type="text/plain")
-            await response(receive, send)
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_POST_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert b'Body: {"hello": "world"}' in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_keepalive(protocol_cls):
-    def app(scope):
-        return Response(b"", status_code=204)
+    app = Response(b"", status_code=204)
 
     protocol = get_connected_protocol(app, protocol_cls)
+    protocol.data_received(SIMPLE_GET_REQUEST)
+    protocol.loop.run_one()
 
+    assert b"HTTP/1.1 204 No Content" in protocol.transport.buffer
+    assert not protocol.transport.is_closing()
+
+
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
+def test_keepalive_timeout(protocol_cls):
+    app = Response(b"", status_code=204)
+
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 204 No Content" in protocol.transport.buffer
     assert not protocol.transport.is_closing()
 
-
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
-def test_keepalive_timeout(protocol_cls):
-    def app(scope):
-        return Response(b"", status_code=204)
-
-    protocol = get_connected_protocol(app, protocol_cls)
-
-    protocol.data_received(SIMPLE_GET_REQUEST)
-    protocol.loop.run_one()
-    assert b"HTTP/1.1 204 No Content" in protocol.transport.buffer
+    protocol.loop.run_later(with_delay=1)
     assert not protocol.transport.is_closing()
 
     protocol.loop.run_later(with_delay=10)
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_close(protocol_cls):
-    def app(scope):
-        return Response(b"", status_code=204, headers={"connection": "close"})
+    app = Response(b"", status_code=204, headers={"connection": "close"})
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
@@ -225,12 +258,11 @@ def test_close(protocol_cls):
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_chunked_encoding(protocol_cls):
-    def app(scope):
-        return Response(
-            b"Hello, world!", status_code=200, headers={"transfer-encoding": "chunked"}
-        )
+    app = Response(
+        b"Hello, world!", status_code=200, headers={"transfer-encoding": "chunked"}
+    )
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
@@ -240,10 +272,36 @@ def test_chunked_encoding(protocol_cls):
     assert not protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
+def test_chunked_encoding_empty_body(protocol_cls):
+    app = Response(
+        b"Hello, world!", status_code=200, headers={"transfer-encoding": "chunked"}
+    )
+
+    protocol = get_connected_protocol(app, protocol_cls)
+    protocol.data_received(SIMPLE_GET_REQUEST)
+    protocol.loop.run_one()
+    assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
+    assert protocol.transport.buffer.count(b"0\r\n\r\n") == 1
+    assert not protocol.transport.is_closing()
+
+
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
+def test_chunked_encoding_head_request(protocol_cls):
+    app = Response(
+        b"Hello, world!", status_code=200, headers={"transfer-encoding": "chunked"}
+    )
+
+    protocol = get_connected_protocol(app, protocol_cls)
+    protocol.data_received(SIMPLE_HEAD_REQUEST)
+    protocol.loop.run_one()
+    assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
+    assert not protocol.transport.is_closing()
+
+
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_pipelined_requests(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
@@ -266,10 +324,9 @@ def test_pipelined_requests(protocol_cls):
     protocol.transport.clear_buffer()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_undersized_request(protocol_cls):
-    def app(scope):
-        return Response(b"xxx", headers={"content-length": "10"})
+    app = Response(b"xxx", headers={"content-length": "10"})
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
@@ -277,10 +334,9 @@ def test_undersized_request(protocol_cls):
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_oversized_request(protocol_cls):
-    def app(scope):
-        return Response(b"xxx" * 20, headers={"content-length": "10"})
+    app = Response(b"xxx" * 20, headers={"content-length": "10"})
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
@@ -288,10 +344,9 @@ def test_oversized_request(protocol_cls):
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_large_post_request(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(LARGE_POST_REQUEST)
@@ -300,33 +355,19 @@ def test_large_post_request(protocol_cls):
     assert not protocol.transport.read_paused
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_invalid_http(protocol_cls):
-    app = lambda scope: None
+    app = Response("Hello, world", media_type="text/plain")
+
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(b"x" * 100000)
+
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_app_exception(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            self.scope = scope
-
-        async def __call__(self, receive, send):
-            raise Exception()
-
-    protocol = get_connected_protocol(App, protocol_cls)
-    protocol.data_received(SIMPLE_GET_REQUEST)
-    protocol.loop.run_one()
-    assert b"HTTP/1.1 500 Internal Server Error" in protocol.transport.buffer
-    assert protocol.transport.is_closing()
-
-
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
-def test_app_init_exception(protocol_cls):
-    def app(scope):
+    async def app(scope, receive, send):
         raise Exception()
 
     protocol = get_connected_protocol(app, protocol_cls)
@@ -336,144 +377,113 @@ def test_app_init_exception(protocol_cls):
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_exception_during_response(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.start", "status": 200})
+        await send({"type": "http.response.body", "body": b"1", "more_body": True})
+        raise Exception()
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.start", "status": 200})
-            await send({"type": "http.response.body", "body": b"1", "more_body": True})
-            raise Exception()
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 500 Internal Server Error" not in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_no_response_returned(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        pass
 
-        async def __call__(self, receive, send):
-            pass
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 500 Internal Server Error" in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_partial_response_returned(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.start", "status": 200})
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.start", "status": 200})
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 500 Internal Server Error" not in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_duplicate_start_message(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.start", "status": 200})
+        await send({"type": "http.response.start", "status": 200})
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.start", "status": 200})
-            await send({"type": "http.response.start", "status": 200})
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 500 Internal Server Error" not in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_missing_start_message(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.body", "body": b""})
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.body", "body": b""})
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 500 Internal Server Error" in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_message_after_body_complete(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.start", "status": 200})
+        await send({"type": "http.response.body", "body": b""})
+        await send({"type": "http.response.body", "body": b""})
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.start", "status": 200})
-            await send({"type": "http.response.body", "body": b""})
-            await send({"type": "http.response.body", "body": b""})
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_value_returned(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        await send({"type": "http.response.start", "status": 200})
+        await send({"type": "http.response.body", "body": b""})
+        return 123
 
-        async def __call__(self, receive, send):
-            await send({"type": "http.response.start", "status": 200})
-            await send({"type": "http.response.body", "body": b""})
-            return 123
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_early_disconnect(protocol_cls):
     got_disconnect_event = False
 
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        nonlocal got_disconnect_event
 
-        async def __call__(self, receive, send):
-            nonlocal got_disconnect_event
+        while True:
+            message = await receive()
+            if message["type"] == "http.disconnect":
+                break
 
-            while True:
-                message = await receive()
-                if message["type"] == "http.disconnect":
-                    break
+        got_disconnect_event = True
 
-            got_disconnect_event = True
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_POST_REQUEST)
     protocol.eof_received()
     protocol.connection_lost(None)
@@ -481,10 +491,9 @@ def test_early_disconnect(protocol_cls):
     assert got_disconnect_event
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_early_response(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(START_POST_REQUEST)
@@ -494,33 +503,30 @@ def test_early_response(protocol_cls):
     assert not protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_read_after_response(protocol_cls):
     message_after_response = None
 
-    class App:
-        def __init__(self, scope):
-            pass
+    async def app(scope, receive, send):
+        nonlocal message_after_response
 
-        async def __call__(self, receive, send):
-            nonlocal message_after_response
+        response = Response("Hello, world", media_type="text/plain")
+        await response(scope, receive, send)
+        message_after_response = await receive()
 
-            response = Response("Hello, world", media_type="text/plain")
-            await response(receive, send)
-            message_after_response = await receive()
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_POST_REQUEST)
     protocol.loop.run_one()
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert message_after_response == {"type": "http.disconnect"}
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_http10_request(protocol_cls):
-    def app(scope):
+    async def app(scope, receive, send):
         content = "Version: %s" % scope["http_version"]
-        return Response(content, media_type="text/plain")
+        response = Response(content, media_type="text/plain")
+        await response(scope, receive, send)
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(HTTP10_GET_REQUEST)
@@ -529,67 +535,86 @@ def test_http10_request(protocol_cls):
     assert b"Version: 1.0" in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_root_path(protocol_cls):
-    def app(scope):
+    async def app(scope, receive, send):
         path = scope.get("root_path", "") + scope["path"]
-        return Response("Path: " + path, media_type="text/plain")
+        response = Response("Path: " + path, media_type="text/plain")
+        await response(scope, receive, send)
 
     protocol = get_connected_protocol(app, protocol_cls, root_path="/app")
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert b"Path: /app/" in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
+def test_raw_path(protocol_cls):
+    async def app(scope, receive, send):
+        path = scope["path"]
+        raw_path = scope.get("raw_path", None)
+        assert "/one/two" == path
+        assert b"/one%2Ftwo" == raw_path
+
+        response = Response("Done", media_type="text/plain")
+        await response(scope, receive, send)
+
+    protocol = get_connected_protocol(app, protocol_cls, root_path="/app")
+    protocol.data_received(GET_REQUEST_WITH_RAW_PATH)
+    protocol.loop.run_one()
+    assert b"Done" in protocol.transport.buffer
+
+
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_max_concurrency(protocol_cls):
-    app = lambda scope: None
+    app = Response("Hello, world", media_type="text/plain")
+
     protocol = get_connected_protocol(app, protocol_cls, limit_concurrency=1)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 503 Service Unavailable" in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_shutdown_during_request(protocol_cls):
-    def app(scope):
-        return Response(b"", status_code=204)
+    app = Response(b"", status_code=204)
 
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.data_received(SIMPLE_GET_REQUEST)
     protocol.shutdown()
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 204 No Content" in protocol.transport.buffer
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_shutdown_during_idle(protocol_cls):
-    app = lambda scope: None
+    app = Response("Hello, world", media_type="text/plain")
+
     protocol = get_connected_protocol(app, protocol_cls)
     protocol.shutdown()
+
     assert protocol.transport.buffer == b""
     assert protocol.transport.is_closing()
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_100_continue_sent_when_body_consumed(protocol_cls):
-    class App:
-        def __init__(self, scope):
-            self.scope = scope
+    async def app(scope, receive, send):
+        body = b""
+        more_body = True
+        while more_body:
+            message = await receive()
+            body += message.get("body", b"")
+            more_body = message.get("more_body", False)
+        response = Response(b"Body: " + body, media_type="text/plain")
+        await response(scope, receive, send)
 
-        async def __call__(self, receive, send):
-            body = b""
-            more_body = True
-            while more_body:
-                message = await receive()
-                body += message.get("body", b"")
-                more_body = message.get("more_body", False)
-            response = Response(b"Body: " + body, media_type="text/plain")
-            await response(receive, send)
-
-    protocol = get_connected_protocol(App, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls)
     EXPECT_100_REQUEST = b"\r\n".join(
         [
             b"POST / HTTP/1.1",
@@ -603,15 +628,15 @@ def test_100_continue_sent_when_body_consumed(protocol_cls):
     )
     protocol.data_received(EXPECT_100_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 100 Continue" in protocol.transport.buffer
     assert b"HTTP/1.1 200 OK" in protocol.transport.buffer
     assert b'Body: {"hello": "world"}' in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_100_continue_not_sent_when_body_not_consumed(protocol_cls):
-    def app(scope):
-        return Response(b"", status_code=204)
+    app = Response(b"", status_code=204)
 
     protocol = get_connected_protocol(app, protocol_cls)
     EXPECT_100_REQUEST = b"\r\n".join(
@@ -627,27 +652,27 @@ def test_100_continue_not_sent_when_body_not_consumed(protocol_cls):
     )
     protocol.data_received(EXPECT_100_REQUEST)
     protocol.loop.run_one()
+
     assert b"HTTP/1.1 100 Continue" not in protocol.transport.buffer
     assert b"HTTP/1.1 204 No Content" in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_unsupported_upgrade_request(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
-    protocol = get_connected_protocol(app, protocol_cls)
+    protocol = get_connected_protocol(app, protocol_cls, ws="none")
     protocol.data_received(UPGRADE_REQUEST)
+
     assert b"HTTP/1.1 400 Bad Request" in protocol.transport.buffer
     assert b"Unsupported upgrade request." in protocol.transport.buffer
 
 
-@pytest.mark.parametrize("protocol_cls", [HttpToolsProtocol, H11Protocol])
+@pytest.mark.parametrize("protocol_cls", HTTP_PROTOCOLS)
 def test_supported_upgrade_request(protocol_cls):
-    def app(scope):
-        return Response("Hello, world", media_type="text/plain")
+    app = Response("Hello, world", media_type="text/plain")
 
-    protocol = get_connected_protocol(app, protocol_cls, ws_protocol_class=WSProtocol)
+    protocol = get_connected_protocol(app, protocol_cls, ws="wsproto")
     protocol.data_received(UPGRADE_REQUEST)
-    assert b"HTTP/1.1 400 Bad Request" in protocol.transport.buffer
-    assert b"Missing Sec-WebSocket-Version header" in protocol.transport.buffer
+
+    assert b"HTTP/1.1 426 " in protocol.transport.buffer
